import { useState, useRef, useMemo, useEffect } from 'react';

var ERR_MODEL = 'model should be a function';
var ERR_PAYLOAD = 'payload should be an object or a function';
var ERR_OUT_MODEL = 'useOutModel passed to get() is not initialized';
var MIGRATE_URL = 'https://github.com/nanxiaobei/flooks#from-v4-to-v5';
var MIGRATE_ERR = "flooks v5 installed, sorry for breaking changes. Simple migrate guide: " + MIGRATE_URL;
var EMPTY_OBJ = {};

var NOOP = function NOOP() {
  return undefined;
};

var __DEV__ = process.env.NODE_ENV !== 'production';

var notObj = function notObj(val) {
  return Object.prototype.toString.call(val) !== '[object Object]';
};

var map = new WeakMap();

var create = function create(model) {
  if (__DEV__ && typeof model !== 'function') throw new Error(ERR_MODEL);

  try {
    useState(0); // eslint-disable-line react-hooks/rules-of-hooks

    throw new Error(MIGRATE_ERR);
  } catch (err) {
    if (err.message === MIGRATE_ERR) throw err;
  }

  var modelSubs = [];
  var modelData = model({
    get: function get(useOutModel) {
      if (typeof useOutModel === 'undefined') return modelData;
      var outData = map.get(useOutModel);
      if (__DEV__ && !outData) throw new Error(ERR_OUT_MODEL);
      return outData;
    },
    set: function set(payload) {
      if (typeof payload === 'function') {
        payload = payload(modelData);
      } else if (__DEV__ && notObj(payload)) {
        throw new Error(ERR_PAYLOAD);
      }

      Object.assign(modelData, payload);
      modelSubs.forEach(function (updater) {
        return updater(payload);
      });
    }
  });

  var useModel = function useModel() {
    var ownData = useRef(EMPTY_OBJ);

    var _useState = useState(false),
        setState = _useState[1];

    var onEffect = useRef(NOOP);
    useMemo(function () {
      var hasState = false;
      var hasUpdate = false;
      var target = {};
      var handler = {
        get: function get(_, key) {
          var val = modelData[key];

          if (typeof val !== 'function') {
            hasState = true;
            target[key] = val;
            return target[key];
          }

          target[key] = new Proxy(val, {
            get: function get(fn, fnKey) {
              if (fnKey === 'loading' && !('loading' in fn)) fn.loading = false;
              return fn[fnKey];
            },
            apply: function apply(fn, _this, args) {
              var res = fn.apply(void 0, args);

              if (!('loading' in fn) || !res || typeof res.then !== 'function') {
                target[key] = fn;
                return res;
              }

              var setLoading = function setLoading(loading) {
                target[key].loading = loading;
                setState(function (s) {
                  return !s;
                });
              };

              target[key] = function () {
                var newRes = fn.apply(void 0, arguments);
                setLoading(true);
                return newRes.finally(function () {
                  return setLoading(false);
                });
              };

              setLoading(true);
              return res.finally(function () {
                return setLoading(false);
              });
            }
          });
          return target[key];
        },
        set: function set(_, key, val) {
          if (key in target) {
            hasUpdate = true;
            target[key] = val;
          }

          return true;
        }
      };
      ownData.current = new Proxy(target, handler);

      onEffect.current = function () {
        if (hasState) {
          handler.get = function (_, key) {
            return target[key];
          };

          var updater = function updater(payload) {
            Object.assign(ownData.current, payload);

            if (hasUpdate) {
              hasUpdate = false;
              setState(function (s) {
                return !s;
              });
            }
          };

          modelSubs.push(updater);
          return function () {
            return modelSubs.splice(modelSubs.indexOf(updater), 1);
          };
        }

        ownData.current = target;
      };
    }, []);
    useEffect(function () {
      return onEffect.current();
    }, []);
    return ownData.current;
  };

  map.set(useModel, modelData);
  return useModel;
};

export { create as default };
